@@ -19,6 +19,8 @@
 #include "shumate-tile-downloader.h"
 #include "shumate-tile-private.h"
 #include "shumate-profiling-private.h"
+#include "shumate-vector-reader.h"
+#include "shumate-vector-reader-iter.h"
 
 /**
  * ShumateVectorRenderer:
@@ -34,6 +36,7 @@
 #include "vector/shumate-vector-symbol-info-private.h"
 #include "vector/shumate-vector-utils-private.h"
 #include "vector/shumate-vector-layer-private.h"
+#include "vector/shumate-vector-index-private.h"
 #endif
 
 struct _ShumateVectorRenderer
@@ -51,6 +54,10 @@ struct _ShumateVectorRenderer
   char *style_json;
 
   GPtrArray *layers;
+
+#ifdef SHUMATE_HAS_VECTOR_RENDERER
+  ShumateVectorIndexDescription *index_description;
+#endif
 };
 
 
@@ -136,6 +143,7 @@ shumate_vector_renderer_finalize (GObject *object)
   g_clear_pointer (&self->source_name, g_free);
   g_clear_object (&self->data_source);
   g_clear_object (&self->sprites);
+  g_clear_pointer (&self->index_description, shumate_vector_index_description_free);
 
   if (self->thread_pool)
     g_thread_pool_free (self->thread_pool, FALSE, FALSE);
@@ -394,18 +402,26 @@ shumate_vector_renderer_initable_init (GInitable     *initable,
         {
           JsonNode *layer_node = json_array_get_element (layers, i);
           JsonObject *layer_obj;
+          const char *layer_id;
           ShumateVectorLayer *layer;
+          ShumateVectorExpression *filter;
 
           if (!shumate_vector_json_get_object (layer_node, &layer_obj, error))
             return FALSE;
 
+          if (!shumate_vector_json_get_string_member (layer_obj, "id", &layer_id, error))
+            return FALSE;
+
           if (!(layer = shumate_vector_layer_create_from_json (layer_obj, error)))
             {
-              g_prefix_error (error, "layer '%s': ", json_object_get_string_member (layer_obj, "id"));
+              g_prefix_error (error, "layer '%s': ", layer_id);
               return FALSE;
             }
 
           g_ptr_array_add (self->layers, layer);
+          filter = shumate_vector_layer_get_filter (layer);
+          if (filter != NULL)
+            shumate_vector_expression_collect_indexes (filter, shumate_vector_layer_get_source_layer (layer), self->index_description);
         }
     }
 
@@ -481,6 +497,9 @@ static void
 shumate_vector_renderer_init (ShumateVectorRenderer *self)
 {
   g_mutex_init (&self->sprites_mutex);
+#ifdef SHUMATE_HAS_VECTOR_RENDERER
+  self->index_description = shumate_vector_index_description_new ();
+#endif
 }
 
 
@@ -592,6 +611,37 @@ shumate_vector_renderer_set_sprite_sheet (ShumateVectorRenderer    *self,
 }
 
 
+/**
+ * shumate_vector_renderer_set_data_source:
+ * @self: a [class@VectorRenderer]
+ * @name: the name of the data source
+ * @data_source: a [class@DataSource]
+ *
+ * Adds a data source to the renderer.
+ *
+ * Currently, [class@VectorRenderer] only supports one data source
+ * and throws an error if the style does not contain exactly one
+ * data source. However, support for multiple sources may be added
+ * in the future, so this method accepts a name parameter. If the
+ * name does not match the one expected by the style, this method
+ * will have no effect.
+ *
+ * Since: 1.2
+ */
+void
+shumate_vector_renderer_set_data_source (ShumateVectorRenderer *self,
+                                         const char            *name,
+                                         ShumateDataSource     *data_source)
+{
+  g_return_if_fail (SHUMATE_IS_VECTOR_RENDERER (self));
+  g_return_if_fail (name != NULL);
+  g_return_if_fail (SHUMATE_IS_DATA_SOURCE (data_source));
+
+  if (g_strcmp0 (name, self->source_name) == 0)
+    g_set_object (&self->data_source, data_source);
+}
+
+
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
 static GdkTexture *
 texture_new_for_surface (cairo_surface_t *surface)
@@ -770,12 +820,11 @@ shumate_vector_renderer_render (ShumateVectorRenderer  *self,
 #ifdef SHUMATE_HAS_VECTOR_RENDERER
   ShumateVectorRenderScope scope;
   cairo_surface_t *surface;
-  gconstpointer data;
-  gsize len;
   g_autoptr(GPtrArray) symbol_list = g_ptr_array_new_with_free_func ((GDestroyNotify)shumate_vector_symbol_info_unref);
   int texture_size;
   g_autofree char *profile_desc = NULL;
   g_autoptr(ShumateVectorSpriteSheet) sprites = NULL;
+  g_autoptr(ShumateVectorReader) reader = NULL;
 
   g_assert (SHUMATE_IS_VECTOR_RENDERER (self));
   g_assert (SHUMATE_IS_TILE (tile));
@@ -794,6 +843,8 @@ shumate_vector_renderer_render (ShumateVectorRenderer  *self,
   scope.zoom_level = shumate_tile_get_zoom_level (tile);
   scope.symbols = symbol_list;
   scope.sprites = sprites;
+  scope.index = NULL;
+  scope.index_description = self->index_description;
 
   if (scope.zoom_level > source_position->zoom)
     {
@@ -817,10 +868,10 @@ shumate_vector_renderer_render (ShumateVectorRenderer  *self,
   scope.cr = cairo_create (surface);
   cairo_scale (scope.cr, scope.scale_factor, scope.scale_factor);
 
-  data = g_bytes_get_data (tile_data, &len);
-  scope.tile = vector_tile__tile__unpack (NULL, len, data);
+  reader = shumate_vector_reader_new (tile_data);
+  scope.reader = shumate_vector_reader_iterate (reader);
 
-  if (scope.tile != NULL)
+  if (scope.reader != NULL)
     for (scope.layer_idx = 0; scope.layer_idx < self->layers->len; scope.layer_idx ++)
       shumate_vector_layer_render ((ShumateVectorLayer *)self->layers->pdata[scope.layer_idx], &scope);
 
@@ -829,7 +880,9 @@ shumate_vector_renderer_render (ShumateVectorRenderer  *self,
 
   cairo_destroy (scope.cr);
   cairo_surface_destroy (surface);
-  vector_tile__tile__free_unpacked (scope.tile, NULL);
+  g_clear_object (&scope.reader);
+
+  g_clear_pointer (&scope.index, shumate_vector_index_free);
 
   profile_desc = g_strdup_printf ("(%d, %d) @ %f", scope.tile_x, scope.tile_y, scope.zoom_level);
   SHUMATE_PROFILE_END (profile_desc);
